#!/usr/bin/env python3
# This script sends feedback to students' forks by opening issues with the Gitlab API

# The target forks and associated metadata are read from stdin. The script expects JSON
# data as outputted by collect_forks.py

# The issues are based on a template, with the content read from a separate CSV table
# with columns "id", "grade", "comments"

import sys
import json
from string import Formatter
from typing import Dict
import requests
import click
import pandas as pd
from teach_utils.common_requests import parse_repo_url


FEEDBACK_TEMPLATE = """
# {project}, group {group}

## Members:
{members}

Please find below your grade and specific comments for this project:

## Grade:
{grade}

## Comments:
{comments}
"""


def fill_template(content: Dict[str, str], template: str) -> str:
    """Replace fields surrounded by {} in template by the value
    of corresponding keys in content.

    Examples
    --------
    >>> content = {'name': 'Bob'}
    >>> fill_template(content, 'I am {name}')
    'I am Bob'
    >>> fill_template(content | {'age': 13}, 'I am {age}')
    'I am 13'
    >>> fill_template(content, 'I am {age}') # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    ValueError: Fields missing from content: ['age']
    """
    try:
        # Replace content in issue template
        return template.format(**content)
    except KeyError:
        templ_flds = [fld for _, fld, _, _ in Formatter().parse(template) if fld]
        missing_flds = [fld for fld in templ_flds if fld not in content.keys()]
        raise ValueError(f"Fields missing from content: {missing_flds}") from None


@click.command()
@click.argument("forks", type=click.File(mode="r"), default=sys.stdin)
@click.option(
    "--token",
    type=click.Path(exists=True),
    help="Path to a file containing the Gitlab API token. If not provided, you will be prompted for the token.",
)
@click.option(
    "--template",
    type=click.Path(exists=True),
    help="Markdown template for the issue. Fields surrounded by {} in the template will be replaced "
    "by content from corresponding columns in feedback.csv. By default, a template with fields {project}, "
    "{group}, {members}, {grade} and {comments} is used.",
    default=None,
)
@click.option(
    "--feedback",
    type=click.Path(exists=True),
    help="CSV file with column 'id', and additional columns matching fields in the template"
    "The id column should match a gitlab project ID from the JSON input.",
    default=None,
)
@click.option(
    "--dry-run",
    is_flag=True,
    help="Print feedback to stdout instead of opening issues"
)
def main(forks, feedback, token, template, dry_run):
    """
    Send feedback to repositories in input JSON by opening issues.
    """

    # Build header with authentication token
    if token is None:
        token = click.prompt(
            "Please enter your Gitlab API token", hide_input=True, err=True
        )
    else:
        token = open(token).read().strip()
    header = {"PRIVATE-TOKEN": token}

    forks = json.load(forks)

    # Load issue template if provided
    if template:
        feedback_template = open(template).read()
    else:
        feedback_template = FEEDBACK_TEMPLATE

    for fork in forks:
        # Get basic fields from repo metadata in json
        base, group, repo = parse_repo_url(fork["url"])
        members = "\n".join([f"* {member['name']}" for member in fork["members"]])
        fill_data = {
            "project": repo.removesuffix(".git"),
            "group": group,
            "members": members,
        }
        # add commit, url and visibility fields from fork data
        fill_data |= {key: fork.get(key) for key in ["url", "commit", "visibility"]}
        # If provided, get additional fields from feedback table
        if feedback:
            feedback_fields = (
                pd.read_csv(feedback)
                .query(f'id == {fork["id"]}')
                .iloc[0]
                .drop("id")
                .to_dict()
            )
            fill_data |= feedback_fields

        # Replace fields in the template by their value for the current fork
        description = fill_template(fill_data, feedback_template)
        # Create issue
        if dry_run:
            print(fork)
            print(description)
        else:
            requests.post(
                f"{base}/api/v4/projects/{fork['id']}/issues?confidential=true&title=Feedback",
                headers=header,
                data={"description": description},
            )


if __name__ == "__main__":
    main()
